
#include <windows.h>
#include <stdio.h>

#include "avisynth.h"
#include "SmoothDeinterlacer15.h"

typedef unsigned char byte;
typedef unsigned __int64 ui64Plane;

class Filter : public GenericVideoFilter
{
private:
  bool   m_fDoubleRate;
  bool   m_fBlend;
  bool   m_fShowDeinterlaced;
  bool   m_fLog;
  int    m_iLaceThresh;
  int	 m_iEdgeThresh;
  int    m_iInterframeLeaveThresh;
  int    m_iInterframeAvg;
  float  m_rLogLacePercent;
  int*   m_piFrameDiffs;
  //changed by Mikola
  int*   m_piFrameDiffsU; //for U-plane
  int*   m_piFrameDiffsV; //for V-plane
  bool	 SSEMMXenabled;
  int	 iDeinterlacedPixels;
  int	 iThreshold;
  int	 iEdgeDetect;
  bool   bShowDeinterlacedAreaOnly;
  bool	 bBlend;
  int	 iInterFrameLeaveThreshold;
  int	 iInterFrameAverage;
  int    iInterFrameAverageRest;
  //end of changed

  HANDLE m_hLog;
  PClip  m_pChildDoubleWeave;
  const char* m_szLogFile;

public:
  Filter(PClip pChildNormal,
         PClip pChildDoubleWeave,
         int   iLaceThresh,
         int   iEdgeThresh,
         int   iInterframeLeaveThresh,
         int   iInterframeAvg,
         bool  fDoubleRate,
         bool  fBlend,
         bool  fShowDeinterlaced,
         bool  fLog,
         float rLogLacePercent,
         const char* szLogFile,
         IScriptEnvironment *pEnv) :
   GenericVideoFilter(pChildNormal),
   m_pChildDoubleWeave(pChildDoubleWeave),
   m_fDoubleRate(fDoubleRate),
   m_fBlend(fBlend),
   m_fShowDeinterlaced(fShowDeinterlaced),
   m_fLog(fLog),
   m_iLaceThresh(iLaceThresh),
   m_iEdgeThresh(iEdgeThresh),
   m_iInterframeLeaveThresh(iInterframeLeaveThresh),
   m_iInterframeAvg(iInterframeAvg),
   m_rLogLacePercent(rLogLacePercent),
   m_szLogFile(szLogFile),
   m_hLog(INVALID_HANDLE_VALUE)

  {
    if (!vi.IsYUY2() && !vi.IsYV12() && !vi.IsRGB32())
		{
			pEnv->ThrowError("SmoothDeinterlace: Supports YUY2, RGB32 and YV12 color formats only");
		}
    if ((vi.width % 4) && vi.IsYV12())
		{
			pEnv->ThrowError("SmoothDeinterlace: Horizontal pixels must divide by 4");
		}

    if(m_fDoubleRate)
    {
      // When we DoubleWeave(), we need to double the frame rate and
      // number of frames.
      vi.fps_numerator *= 2;
      vi.num_frames *= 2;
    }

    // We always output frames
    vi.SetFieldBased(false);

    // Get a place to put the interframe differences
    int iInterframeLeaveThreshX=(m_iInterframeLeaveThresh *
                           m_iInterframeLeaveThresh) + 1;    
	m_piFrameDiffs = new int[vi.width * vi.height];
	
	if (vi.IsYV12())
		{
			m_piFrameDiffsU = new int[(vi.width * vi.height)>>2];
			m_piFrameDiffsV = new int[(vi.width * vi.height)>>2];
		    for(int i = 0; i < (vi.width * vi.height); i+=2)
			{
				m_piFrameDiffs[i] = iInterframeLeaveThreshX;
				m_piFrameDiffs[i+1] = iInterframeLeaveThreshX;
			}

			for(i = 0; i < ((vi.width * vi.height)>>2); i+=1)
			{
					m_piFrameDiffsU[i] = iInterframeLeaveThreshX;	
					m_piFrameDiffsV[i] = iInterframeLeaveThreshX;
			}
		}
	else
		{
        	for (int i = 0; i < (vi.width * vi.height); i+=1)		
				m_piFrameDiffs[i] = iInterframeLeaveThreshX;
		} 


    // Open the log file if necessary
    if(m_fLog)
    {
      m_hLog = CreateFile(m_szLogFile,
                          GENERIC_WRITE,
                          FILE_SHARE_READ,
                          NULL,
                          CREATE_ALWAYS,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);
      if(m_hLog != INVALID_HANDLE_VALUE)
      {
        DWORD dw;
        WriteFile(m_hLog,
                  "Frame\tInterlaced area (%)\r\n",
                  27,
                  &dw,
                  NULL);
      }
    }
  }

  Filter::~Filter()
  {
    if(m_hLog != INVALID_HANDLE_VALUE)
    {
      CloseHandle(m_hLog);
    }
    delete [] m_piFrameDiffs;
	if (vi.IsYV12())
	{
	delete [] m_piFrameDiffsU;
	delete [] m_piFrameDiffsV;
	}
	
  }

  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* pEnv);
  int __stdcall ProcessFrame_YV12_SSEMMX(int n, ui64Plane* pframe, ui64Plane* pframeprev, ui64Plane* pdst, int* piFrameDiffs, const int src_w, int src_h, int src_p, int dst_p);
  int __stdcall ProcessFrame_YV12(int n, byte* pframe, byte* pframeprev, byte* pdst, int* piFrameDiffs, const int src_w, int src_h, int src_p, int dst_p);

};


PVideoFrame __stdcall Filter::GetFrame(int  n,  IScriptEnvironment* pEnv)
{
  PVideoFrame pvfCurrent;
  PVideoFrame pvfPrev;
  PVideoFrame pvfNew;

  iDeinterlacedPixels = 0;

// Determine which of the potential input streams is needed.
  PClip pChild;
  if(m_fDoubleRate)
    pChild = m_pChildDoubleWeave;
  else
    pChild = child;


  if(m_fDoubleRate)
  {
    // A trick here to get the frames aligned correctly.
    if(n)
      --n;
  }

  pvfCurrent = pChild->GetFrame(n, pEnv);

  if(n > 1)
    pvfPrev = pChild->GetFrame(n - 2, pEnv);
  else
    pvfPrev = pvfCurrent;


  SSEMMXenabled = (pEnv->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;


  iThreshold = m_iLaceThresh;
  iThreshold = iThreshold * iThreshold * 4;
  iEdgeDetect = m_iEdgeThresh;
  if(iEdgeDetect > 180)
    iEdgeDetect = 180;  // We don't want an integer overflow in the interlace calculation.
  iEdgeDetect = iEdgeDetect * iEdgeDetect;
  bShowDeinterlacedAreaOnly = m_fShowDeinterlaced;
  bBlend = m_fBlend;
  iInterFrameLeaveThreshold = m_iInterframeLeaveThresh *
                                  m_iInterframeLeaveThresh;
  iInterFrameAverage = m_iInterframeAvg * 1024 / 100;
  if(iInterFrameAverage < 0)
    iInterFrameAverage = 0;
  if(iInterFrameAverage > 1023)
    iInterFrameAverage = 1023;
  iInterFrameAverageRest = 1024 - iInterFrameAverage;

  const int iSrcPitch = pvfCurrent->GetPitch();

  if(vi.IsRGB32())
  {
    typedef unsigned long Pixel32;

    int iR0, iG0, iB0, iR1, iG1, iB1, iR2, iG2, iB2, iR3, iG3, iB3;
    int iR2p, iG2p, iB2p;
    Pixel32* pframe1;
    Pixel32* pframe2;
    Pixel32* pframe3;
    Pixel32* pdst1;
    Pixel32* pframe2p;
    int* piFrameDiffs2p;
    int iInterlaceValue0, iInterlaceValue1, iInterlaceValue2;
    int iFrameDiffValue0, iFrameDiffValue1, iFrameDiffValue2;
    iR1 = iG1 = iB1 = 0;  // Avoid compiler warning. The value is not used.

    Pixel32* pframe = (Pixel32*)pvfCurrent->GetReadPtr();
    Pixel32* pframeprev = (Pixel32*)pvfPrev->GetReadPtr();
    pvfNew = pEnv->NewVideoFrame(vi);
    Pixel32* pdst = (Pixel32*)pvfNew->GetWritePtr();
    int* piFrameDiffs = m_piFrameDiffs;

    const int iDstPitch = pvfNew->GetPitch();

    int iOddEven;
    if(m_fDoubleRate)
      iOddEven = (n & 1);
    else
      iOddEven = 1;

    // Do some area based deinterlacing on the combined frame.
    for(int x = 0; x < vi.width; x++)
    {
      pframe3 = pframe + x;
      iR3 = (*pframe3 >> 16) & 0xff;
      iG3 = (*pframe3 >> 8) & 0xff;
      iB3 = *pframe3 & 0xff;
      pframe2 = (Pixel32 *)((char *)pframe3 + iSrcPitch);
      iR2 = (*pframe2 >> 16) & 0xff;
      iG2 = (*pframe2 >> 8) & 0xff;
      iB2 = *pframe2 & 0xff;
    
      pframe2p = pframeprev + x;
      piFrameDiffs2p = piFrameDiffs + (x * vi.height);

      pdst1 = pdst + x;
      iInterlaceValue1 = iInterlaceValue2 = 0;
      iFrameDiffValue1 = iFrameDiffValue2 = 0;

      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
      for(int y = 0; y <= vi.height; y++)
      {
        pframe1 = pframe2;
        pframe2 = pframe3;
        pframe3 = (Pixel32 *)((char *)pframe3 + iSrcPitch);
        iR0 = iR1;
        iG0 = iG1;
        iB0 = iB1;
        iR1 = iR2;
        iG1 = iG2;
        iB1 = iB2;
        iR2 = iR3;
        iG2 = iG3;
        iB2 = iB3;
        if(y < vi.height - 1)
        {
          iR3 = (*pframe3 >> 16) & 0xff;
          iG3 = (*pframe3 >> 8) & 0xff;
          iB3 = *pframe3 & 0xff;

          iR2p = (*pframe2p >> 16) & 0xff;
          iG2p = (*pframe2p >> 8) & 0xff;
          iB2p = *pframe2p & 0xff;
        }
        else
        {
          iR3 = iR1;
          iG3 = iG1;
          iB3 = iB1;
        }
        pframe2p = (Pixel32 *)((char *)pframe2p + iSrcPitch);

        iInterlaceValue0 = iInterlaceValue1;
        iInterlaceValue1 = iInterlaceValue2;
        iFrameDiffValue0 = iFrameDiffValue1;
        iFrameDiffValue1 = iFrameDiffValue2;
        if(y < vi.height)
        {
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
          // 3, 6 and 1 are approximate values for converting R, G, B to intensity.
          iInterlaceValue2 = (3 * ((iR1 - iR2) * (iR3 - iR2) - ((iEdgeDetect * (iR1 - iR3) * (iR1 - iR3)) >> 12)) +
                    6 * ((iG1 - iG2) * (iG3 - iG2) - ((iEdgeDetect * (iG1 - iG3) * (iG1 - iG3)) >> 12)) +
                    (iB1 - iB2) * (iB3 - iB2) - ((iEdgeDetect * (iB1 - iB3) * (iB1 - iB3)) >> 12));
          iFrameDiffValue2 = (iInterFrameAverage * *piFrameDiffs2p +
                        iInterFrameAverageRest * (n > 1 ?
                          3 * (iR2 - iR2p) * (iR2 - iR2p) + 6 * (iG2 - iG2p) * (iG2 - iG2p) + (iB2 - iB2p) * (iB2 - iB2p) :
                          *piFrameDiffs2p)) >> 10;
          *piFrameDiffs2p = iFrameDiffValue2;
          piFrameDiffs2p++;
        }
        else
        {
          iInterlaceValue2 = 0;
          iFrameDiffValue2 = 0;
        }

        if(y > 0)
        {  // Get mean interlace value of 3 rows (i.e. 5 pixels are examined).
          // Middle row has twice the weight.
          if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold &&
            iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
          {
            if(bShowDeinterlacedAreaOnly)
            {
              *pdst1 = 0x00903030 + (bBlend ? (((iR0 + 2 * iR1 + iR2) >> 4) << 16) + (((iG0 + 2 * iG1 + iG2) >> 4) << 8) + ((iB0 + 2 * iB1 + iB2) >> 4)
                                  :
                                    (y % 2 == iOddEven ? (iR1 >> 2 << 16) + (iG1 >> 2 << 8) + (iB1 >> 2) : (((iR0 + iR2) >> 3) << 16) + (((iG0 + iG2) >> 3) << 8) + ((iB0 + iB2) >> 3)));
            }
            else
            {
              // Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
              // Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
              *pdst1 = bBlend ? (((iR0 + 2 * iR1 + iR2) >> 2) << 16) + (((iG0 + 2 * iG1 + iG2) >> 2) << 8) + ((iB0 + 2 * iB1 + iB2) >> 2)
                        :
                        (y % 2 == iOddEven ? *pframe1 : (((iR0 + iR2) >> 1) << 16) + (((iG0 + iG2) >> 1) << 8) + ((iB0 + iB2) >> 1));

            }
            iDeinterlacedPixels++;
          }
          else
          {
            if(bShowDeinterlacedAreaOnly)
            {
              *pdst1 = (iR1 >> 2 << 16) + (iG1 >> 2 << 8) + (iB1 >> 2);
              if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold)
                *pdst1 += 0x00309030;
              else if(iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
                *pdst1 += 0x00303090;
              else
                *pdst1 += 0x00303030;
            }
            else
              *pdst1 = *pframe1;
          }
          pdst1 = (Pixel32 *)((char *)pdst1 + iDstPitch);
        }
      }
    }
  }
  else if(vi.IsYUY2())
  {
    struct PixelPair
    {
      BYTE Y0;
      BYTE U;
      BYTE Y1;
      BYTE V;
    };

    PixelPair PixelPair0, PixelPair1 = {0, 0, 0, 0}, PixelPair2, PixelPair3;
    int iY2pe, iY2po;
    PixelPair *pframe1, *pframe2, *pframe3, *pdst1, *pframe2p;
    int *piFrameDiffs2p;
    int iInterlaceValue0e, iInterlaceValue1e, iInterlaceValue2e;
    int iInterlaceValue0o, iInterlaceValue1o, iInterlaceValue2o;
    int iFrameDiffValue0e, iFrameDiffValue1e, iFrameDiffValue2e;
    int iFrameDiffValue0o, iFrameDiffValue1o, iFrameDiffValue2o;

    PixelPair* pframe = (PixelPair*)pvfCurrent->GetReadPtr();
    PixelPair* pframeprev = (PixelPair*)pvfPrev->GetReadPtr();
    pvfNew = pEnv->NewVideoFrame(vi);
    PixelPair* pdst = (PixelPair*)pvfNew->GetWritePtr();
    int* piFrameDiffs = m_piFrameDiffs;

    const int iDstPitch = pvfNew->GetPitch();

    int iOddEven;
    if(m_fDoubleRate)
      iOddEven = (n & 1);
    else
      iOddEven = 1;

    // Do some area based deinterlacing on the combined frame.
    for(int x = 0; x < vi.width; x += 2)
    {
      pframe3 = pframe + (x >> 1);
      PixelPair3 = *pframe3;
      pframe2 = (PixelPair*)((DWORD)pframe3 + iSrcPitch);
      PixelPair2 = *pframe2;
    
      pframe2p = pframeprev + (x >> 1);
      piFrameDiffs2p = piFrameDiffs + (x * vi.height);

      pdst1 = pdst + (x >> 1);
      iInterlaceValue1e = iInterlaceValue2e = 0;
      iInterlaceValue1o = iInterlaceValue2o = 0;
      iFrameDiffValue1e = iFrameDiffValue2e = 0;
      iFrameDiffValue1o = iFrameDiffValue2o = 0;

      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
      for(int y = 0; y <= vi.height; y++)
      {
        pframe1 = pframe2;
        pframe2 = pframe3;
        pframe3 = (PixelPair*)((DWORD)pframe3 + iSrcPitch);
        PixelPair0 = PixelPair1;
        PixelPair1 = PixelPair2;
        PixelPair2 = PixelPair3;
        if(y < vi.height - 1)
        {
          PixelPair3 = *pframe3;
          iY2pe = *((unsigned char*)pframe2p);
          iY2po = *((unsigned char*)pframe2p + 2);
        }
        else
        {
          PixelPair3 = PixelPair1;
        }
        pframe2p = (PixelPair*)((DWORD)pframe2p + iSrcPitch);

        iInterlaceValue0e = iInterlaceValue1e;
        iInterlaceValue0o = iInterlaceValue1o;
        iInterlaceValue1e = iInterlaceValue2e;
        iInterlaceValue1o = iInterlaceValue2o;
        iFrameDiffValue0e = iFrameDiffValue1e;
        iFrameDiffValue0o = iFrameDiffValue1o;
        iFrameDiffValue1e = iFrameDiffValue2e;
        iFrameDiffValue1o = iFrameDiffValue2o;
        if(y < vi.height)
        {
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
          iInterlaceValue2e = 10 * ((PixelPair1.Y0 - PixelPair2.Y0) * (PixelPair3.Y0 - PixelPair2.Y0) - ((iEdgeDetect * (PixelPair1.Y0 - PixelPair3.Y0) * (PixelPair1.Y0 - PixelPair3.Y0)) >> 12));
          iInterlaceValue2o = 10 * ((PixelPair1.Y1 - PixelPair2.Y1) * (PixelPair3.Y1 - PixelPair2.Y1) - ((iEdgeDetect * (PixelPair1.Y1 - PixelPair3.Y1) * (PixelPair1.Y1 - PixelPair3.Y1)) >> 12));
          iFrameDiffValue2e = (iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest *
                              (n                       ?
                               10 * (PixelPair2.Y0 - iY2pe) * (PixelPair2.Y0 - iY2pe) :
                               *piFrameDiffs2p)) >> 10;
          iFrameDiffValue2o = (iInterFrameAverage * *(piFrameDiffs2p + 1) + iInterFrameAverageRest *
                              (n                       ?
                               10 * (PixelPair2.Y1 - iY2po) * (PixelPair2.Y1 - iY2po) :
                               *(piFrameDiffs2p + 1))) >> 10;
          *piFrameDiffs2p = iFrameDiffValue2e;
          *(piFrameDiffs2p + 1) = iFrameDiffValue2o;
          piFrameDiffs2p++;
        }
        else
        {
          iInterlaceValue2e = 0;
          iInterlaceValue2o = 0;
          iFrameDiffValue2e = 0;
          iFrameDiffValue2o = 0;
        }

        if(y > 0)
        {
          // Get mean interlace value of 3 rows (i.e. 5 pixels are examined).
          // Middle row has twice the weight.
          if(iInterlaceValue0e + 2 * iInterlaceValue1e + iInterlaceValue2e > iThreshold &&
             iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
          {
            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold &&
               iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
            {
              // Even and odd are both interlaced
              if(bBlend)
              {
                // Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = (PixelPair0.Y0 + 2 * PixelPair1.Y0 + PixelPair2.Y0) / 4;
                PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U  + PixelPair2.U)  / 4;
                PixelPairDst.Y1 = (PixelPair0.Y1 + 2 * PixelPair1.Y1 + PixelPair2.Y1) / 4;
                PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V  + PixelPair2.V)  / 4;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                // Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = (PixelPair0.Y0 + PixelPair2.Y0) / 2;
                  PixelPairDst.U  = (PixelPair0.U  + PixelPair2.U)  / 2;
                  PixelPairDst.Y1 = (PixelPair0.Y1 + PixelPair2.Y1) / 2;
                  PixelPairDst.V  = (PixelPair0.V  + PixelPair2.V)  / 2;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels += 2;
            }
            else
            {
              // Even is interlaced, but odd is not
              if(bBlend)
              {
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = (PixelPair0.Y0 + 2 * PixelPair1.Y0 + PixelPair2.Y0) / 4;
                PixelPairDst.U  = (PixelPair0.U  + 6 * PixelPair1.U  + PixelPair2.U)  / 8;
                PixelPairDst.Y1 = PixelPair1.Y1;
                PixelPairDst.V  = (PixelPair0.V  + 6 * PixelPair1.V  + PixelPair2.V)  / 8;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = (PixelPair0.Y0 + PixelPair2.Y0) / 2;
                  PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U + PixelPair2.U)  / 4;
                  PixelPairDst.Y1 = PixelPair1.Y1;
                  PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V + PixelPair2.V)  / 4;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels++;
            }
          }
          else
          {
            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold &&
               iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
            {
              // Even is not interlaced, but odd is
              if(bBlend)
              {
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = PixelPair1.Y0;
                PixelPairDst.U  = (PixelPair0.U  + 6 * PixelPair1.U  + PixelPair2.U)  / 8;
                PixelPairDst.Y1 = (PixelPair0.Y1 + 2 * PixelPair1.Y1 + PixelPair2.Y1) / 4;
                PixelPairDst.V  = (PixelPair0.V  + 6 * PixelPair1.V  + PixelPair2.V)  / 8;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = PixelPair1.Y0;
                  PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U + PixelPair2.U)  / 4;
                  PixelPairDst.Y1 = (PixelPair0.Y1 + PixelPair2.Y1) / 2;
                  PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V + PixelPair2.V)  / 4;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels++;
            }
            else
            {
              // Neither even nor odd is interlaced
              *pdst1 = *pframe1;
            }
          }

          if(bShowDeinterlacedAreaOnly)
          {
            int iY0 = pdst1->Y0 / 4;
            int iU0 = (pdst1->U - 128) / 4;
            int iU1 = iU0;
            int iY1 = pdst1->Y1 / 4;
            int iV0 = (pdst1->V - 128) / 4;
            int iV1 = iV0;

            if(iInterlaceValue0e + 2 * iInterlaceValue1e + iInterlaceValue2e > iThreshold)
            {
              if(iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
              {
                iY0 += 76;
                iU0 += -14;
                iV0 += 41;
              }
              else
              {
                iY0 += 103;
                iU0 += -28;
                iV0 += -34;
              }
            }
            else
            {
              if(iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
              {
                iY0 += 58;
                iU0 += 41;
                iV0 += -6;
              }
              else
              {
                iY0 += 47;
              }
            }

            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold)
            {
              if(iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
              {
                iY1 += 76;
                iU1 += -14;
                iV1 += 41;
              }
              else
              {
                iY1 += 103;
                iU1 += -28;
                iV1 += -34;
              }
            }
            else
            {
              if(iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
              {
                iY1 += 58;
                iU1 += 41;
                iV1 += -6;
              }
              else
              {
                iY1 += 47;
              }
            }

            pdst1->Y0 = iY0;
            pdst1->U = ((iU0 + iU1) / 2) + 128;
            pdst1->Y1 = iY1;
            pdst1->V = ((iV0 + iV1) / 2) + 128;
          }

          pdst1 = (PixelPair*)((DWORD)pdst1 + iDstPitch);
        }
      }
    }
  }
  else if(vi.IsYV12())
  {
   int i;

   int* piFrameDiffs = m_piFrameDiffs;
   int* piFrameDiffsU = m_piFrameDiffsU;
   int* piFrameDiffsV = m_piFrameDiffsV;
   
    pvfNew = pEnv->NewVideoFrame(vi); //destination frame
    const int iDstPitch = pvfNew->GetPitch(); // Requests pitch (length of a line) of the destination image.
   
    const int src_width = pvfCurrent->GetRowSize();
    const int src_height = pvfCurrent->GetHeight();

    const int iSrcPitchUV = pvfCurrent->GetPitch(PLANAR_U);	// The pitch,height and width information
    const int src_widthUV = pvfCurrent->GetRowSize(PLANAR_U);	// is guaranted to be the same for both
    const int src_heightUV = pvfCurrent->GetHeight(PLANAR_U);	// the U and V planes so we only the U
    const int iDstPitchUV = pvfNew->GetPitch(PLANAR_U);		// plane values and use them for V

   if  (SSEMMXenabled && !bShowDeinterlacedAreaOnly && !m_fLog)
   { 
   //begin to process Y-plane
     ui64Plane* pframe = (ui64Plane*)pvfCurrent->GetReadPtr();
     ui64Plane* pframeprev = (ui64Plane*)pvfPrev->GetReadPtr();
     ui64Plane* pdst = (ui64Plane*)pvfNew->GetWritePtr();

      i=ProcessFrame_YV12_SSEMMX(n, pframe, pframeprev, pdst, piFrameDiffs, src_width, src_height, iSrcPitch, iDstPitch);
     //begin to process U-plane
     pframe = (ui64Plane*)pvfCurrent->GetReadPtr(PLANAR_U);
     pframeprev = (ui64Plane*)pvfPrev->GetReadPtr(PLANAR_U);
     pdst = (ui64Plane*)pvfNew->GetWritePtr(PLANAR_U);
     
      i=ProcessFrame_YV12_SSEMMX(n, pframe, pframeprev, pdst, piFrameDiffsU, src_widthUV, src_heightUV, iSrcPitchUV, iDstPitchUV);

     //begin to process V-plane
     pframe = (ui64Plane*)pvfCurrent->GetReadPtr(PLANAR_V);
     pframeprev = (ui64Plane*)pvfPrev->GetReadPtr(PLANAR_V);
     pdst = (ui64Plane*)pvfNew->GetWritePtr(PLANAR_V);

      i=ProcessFrame_YV12_SSEMMX(n, pframe, pframeprev, pdst, piFrameDiffsV, src_widthUV, src_heightUV, iSrcPitchUV, iDstPitchUV);
	  __asm emms;
    }
   else
   {
   //begin to process Y-plane
     byte* pframe = (byte*)pvfCurrent->GetReadPtr();
     byte* pframeprev = (byte*)pvfPrev->GetReadPtr();
     byte* pdst = (byte*)pvfNew->GetWritePtr();

      i=ProcessFrame_YV12(n, pframe, pframeprev, pdst, piFrameDiffs, src_width, src_height, iSrcPitch, iDstPitch);

     //begin to process U-plane
     pframe = (byte*)pvfCurrent->GetReadPtr(PLANAR_U);
     pframeprev = (byte*)pvfPrev->GetReadPtr(PLANAR_U);
     pdst = (byte*)pvfNew->GetWritePtr(PLANAR_U);
	 bShowDeinterlacedAreaOnly=false;

      i=ProcessFrame_YV12(n, pframe, pframeprev, pdst, piFrameDiffsU, src_widthUV, src_heightUV, iSrcPitchUV, iDstPitchUV);

     //begin to process V-plane
     pframe = (byte*)pvfCurrent->GetReadPtr(PLANAR_V);
     pframeprev = (byte*)pvfPrev->GetReadPtr(PLANAR_V);
     pdst = (byte*)pvfNew->GetWritePtr(PLANAR_V);

      i=ProcessFrame_YV12(n, pframe, pframeprev, pdst, piFrameDiffsV, src_widthUV, src_heightUV, iSrcPitchUV, iDstPitchUV);
   }
 }

 else
    pEnv->ThrowError("SmoothDeinterlace: only RGB32, YUY2 and YV12 formats supported!");

  if(m_fLog)
  {
    float fDeinterlacedPercent = 100.0f * iDeinterlacedPixels / vi.width / vi.height;
    if(fDeinterlacedPercent > m_rLogLacePercent && m_hLog != INVALID_HANDLE_VALUE)
    {
      char szLog[25];
      sprintf(szLog, "%li\t%5.1f\r\n", n, fDeinterlacedPercent);
      DWORD dw;
      WriteFile(m_hLog, szLog, strlen(szLog), &dw, NULL);
    }
  }
  
  return(pvfNew);
}

int __stdcall Filter::ProcessFrame_YV12_SSEMMX(int n, ui64Plane* pframe, ui64Plane* pframeprev, ui64Plane* pdst, int* piFrameDiffs, const int src_w, int src_h, int src_p, int dst_p)
{

     __int64 i64InterlaceValue0lol, i64InterlaceValue0loh, i64InterlaceValue0hil, i64InterlaceValue0hih, 
                i64InterlaceValue1lol, i64InterlaceValue1loh, i64InterlaceValue1hil, i64InterlaceValue1hih, 
                i64InterlaceValue2lol, i64InterlaceValue2loh, i64InterlaceValue2hil, i64InterlaceValue2hih; 
	 __int64 i64FrameDiffValue0lol, i64FrameDiffValue0loh, i64FrameDiffValue0hil, i64FrameDiffValue0hih,
				i64FrameDiffValue1lol, i64FrameDiffValue1loh, i64FrameDiffValue1hil, i64FrameDiffValue1hih,
				i64FrameDiffValue2lol, i64FrameDiffValue2loh, i64FrameDiffValue2hil, i64FrameDiffValue2hih;
     __int64  i64EdgeDetect, i64InterFrameAverage, i64InterFrameAverageRest, i64InterFrameLeaveThreshold,
               i64Threshold, 	i64SavelumaLo, i64SavelumaHi;

     ui64Plane  i64widthYV12_dst_p;
     ui64Plane  i64iWidth_rest_src_w;
     ui64Plane  i64pframe_pframeprev;
     ui64Plane  i64pdst_piFrameDiffs;
     ui64Plane  i64pframe1_pdst1;
     ui64Plane  i64pframe3_2;
     ui64Plane  i64pframe2p_piFrameDiffs2p;
     ui64Plane  bY2p;
	int iThresholdYV12=Filter::iThreshold;
	int iEdgeDetectYV12=Filter::iEdgeDetect;
	int iInterFrameAverageYV12=Filter::iInterFrameAverage;
	int iInterFrameAverageRestYV12=Filter::iInterFrameAverageRest;
	int iInterFrameLeaveThresholdYV12=Filter::iInterFrameLeaveThreshold;
	int	widthYV12;
	int iWidth_rest;
	short shHeightYV12;
	bool bN;
	bool bBlendYV12;
	bool bOddEven;


    shHeightYV12 = (short)src_h;
    iWidth_rest = src_w % 8;
	widthYV12 = src_w - iWidth_rest;  
    
	if(Filter::m_fDoubleRate)
      bOddEven = (n & 1);
    else
      bOddEven = 1;

    if (n==0)
      bN=true;
    else
      bN=false;


    bBlendYV12 = Filter::bBlend;

		i64InterlaceValue2lol = i64InterlaceValue1lol = i64InterlaceValue0lol = 0;
		i64InterlaceValue2loh = i64InterlaceValue1loh = i64InterlaceValue0loh = 0;
		i64InterlaceValue2hil = i64InterlaceValue1hil = i64InterlaceValue0hil = 0;
		i64InterlaceValue2hih = i64InterlaceValue1hih = i64InterlaceValue0hih = 0;
		i64FrameDiffValue2lol = i64FrameDiffValue1lol = i64FrameDiffValue0lol = 0;
		i64FrameDiffValue2loh = i64FrameDiffValue1loh = i64FrameDiffValue0loh = 0;
		i64FrameDiffValue2hil = i64FrameDiffValue1hil = i64FrameDiffValue0hil = 0;
		i64FrameDiffValue2hih = i64FrameDiffValue1hih = i64FrameDiffValue0hih = 0;
   	
  // Do some area based deinterlacing on the plane.
	__asm
	{
		pxor	        mm1, mm1            //bPlane1=0
        mov             eax, widthYV12
        mov             ebx, dst_p
        movd            mm2, eax            //widthYV12
        movd            mm3, ebx            //dst_p
        psllq           mm2, 32
        por             mm3, mm2
        movntq          i64widthYV12_dst_p, mm3
        mov             eax, iWidth_rest
        mov             ebx, src_w
        movd            mm2, eax            //iWidth_rest
        movd            mm3, ebx            //src_w
        psllq           mm2, 32
        por             mm3, mm2
        movntq          i64iWidth_rest_src_w, mm3
        mov             eax, pframe
        mov             ebx, pframeprev
        movd            mm2, eax            //pframe
        movd            mm3, ebx            //dst_p
        psllq           mm2, 32
        por             mm3, mm2
        movq            i64pframe_pframeprev, mm3
        mov             eax, pdst
        mov             ebx, piFrameDiffs
        movd            mm2, eax            //pdst
        movd            mm3, ebx            //piFrameDiffs
        psllq           mm2, 32
        por             mm3, mm2
        movq            i64pdst_piFrameDiffs, mm3

		mov				esi, iEdgeDetectYV12  //load iEdgeDetect
		mov				edi, iInterFrameAverageYV12  //load iInterFrameAverage
		mov				ebx, iInterFrameAverageRestYV12  //load iInterFrameAverageRest
		mov				eax, iInterFrameLeaveThresholdYV12
		pinsrw	        mm5, esi,0
		pinsrw	        mm5, esi,1
		pinsrw	        mm5, esi,2
		pinsrw	        mm5, esi,3
		movntq	        i64EdgeDetect, mm5
        pxor            mm5, mm5
		pinsrw	        mm5, edi,0
		pinsrw	        mm5, edi,2
		movntq	        i64InterFrameAverage, mm5
		pinsrw	        mm5, ebx,0
		pinsrw	        mm5, ebx,1
		pinsrw	        mm5, ebx,2
		pinsrw	        mm5, ebx,3
		movntq	        i64InterFrameAverageRest, mm5
		pxor			mm5, mm5
		pinsrw	        mm5, eax,0
		pinsrw	        mm5, eax,2
		movntq 	        i64InterFrameLeaveThreshold, mm5
		mov				esi, iThresholdYV12
		pxor			mm5, mm5
		pinsrw	        mm5, esi,0
		pinsrw	        mm5, esi,2
		movntq	        i64Threshold, mm5

		// Loop general reg usage
		//
		// eax -  
		// ebx - src 2 lines above curr 
		// ecx loop ctr
		// edx - dst_pit
		// edi - dest
		// esi - src pixels, 1 line up

		// now loop and get the middle qwords

        xor			    eax, eax
        xor             ebx, ebx
		xor			    ecx, ecx
        mov             bl, bOddEven
        shl             ebx, 16
		mov             bl, bN
   		mov             bh, bBlendYV12
        mov			    edx, src_p
        mov             cx, shHeightYV12
        shl             ecx, 16
LoopW:  
        ror             ebx, 16
        mov             bh, 1                   // for check if current y = 0 
        ror             ebx, 16
        mov             esi, ecx
        shr             esi, 16
        xor             cx, cx
        add             ecx, esi                //restore in cx src_h
		movq		    mm4, i64pframe_pframeprev //pframe - in high, pframeprev - low order dwords
		movq		    mm5, mm4
		psrlq		    mm5, 32
		movd		    esi, mm5			//mov	esi, pframe  -	current row, current frame
		add			    esi, eax			// pframe3 = pframe + x;
		movq            mm3, qword ptr[esi] // in mm3 - 8 bytes from current line, current frame (bPlane3 = *pframe3)
        lea             edi, [esi+edx]		// pframe2 = pframe3 + src_p; 
		movq	        mm2, qword ptr[edi] // in mm2 - 8 bytes from next line, current frame (bPlane2 = *pframe2)
        movd            mm5, esi            //pframe3    
        movd            mm6, edi            //pframe2
        psllq           mm5, 32
        por             mm5, mm6
        movq            i64pframe3_2, mm5   // pframe3 - high, pframe2 - in low order dwords
        movd            mm5, eax            //load x
        psllq           mm5, 32
        movd            mm6, eax
        por             mm5, mm6

   		movq		    mm6, mm4            //load with pframeprev
        psllq           mm6, 32
        movq            mm4, i64pdst_piFrameDiffs
        movq            mm7, mm4
        psrlq           mm7, 32             //load with pdst
        por             mm6, mm7
        paddd           mm6, mm5            // pframe2p = pframeprev + x - in high order dword
                                            // pdst1 = pdst + x          - in low order dword
        movq            i64pframe1_pdst1, mm6  //save pdst1, "pframe1" could be ignored 
        pxor            mm5, mm5
        mov             esi, ecx
        shr             esi, 16             
        imul            esi, eax            // src_h * eax
        movd            mm7, esi            // load src_h * eax
        punpckldq       mm4, mm5            // load piFrameDiffs
        paddd           mm7, mm4            // piFrameDiffs2p=piFrameDiffs + x * src_h - in low order dword
        punpckhdq       mm5, mm6            // load pframe2p
        por             mm7, mm5
        movq            i64pframe2p_piFrameDiffs2p, mm7
		
		pxor		    mm4, mm4
		movq		    i64FrameDiffValue2hih, mm4
		movq		    i64FrameDiffValue1hih, mm4
   		movq	    	i64InterlaceValue2hih, mm4
		movq		    i64InterlaceValue1hih, mm4
        movq		    i64FrameDiffValue2hil, mm4
        movq		    i64FrameDiffValue2loh, mm4
        movq		    i64FrameDiffValue2lol, mm4
        movq		    i64FrameDiffValue1hil, mm4
        movq		    i64FrameDiffValue1loh, mm4
        movq		    i64FrameDiffValue1lol, mm4
   		movq		    i64InterlaceValue2hil, mm4
		movq		    i64InterlaceValue2loh, mm4
		movq		    i64InterlaceValue2lol, mm4
        movq		    i64InterlaceValue1hil, mm4
        movq		    i64InterlaceValue1loh, mm4
        movq		    i64InterlaceValue1lol, mm4
      
      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
	  //for(int y = 0; y <= src_height; y++)

Loop_H:	
        movq            mm4, i64pframe3_2
        movq            mm6, mm4
        psllq           mm6, 32              // pframe2 - in high dword                                                                
        movq            mm5, i64pframe1_pdst1
        psllq           mm5, 32
        psrlq           mm5, 32              //load pdst1 in low dword
        por             mm6, mm5             // pframe1 = pframe2
        movq            i64pframe1_pdst1, mm6 // store pframe1 = pframe2, pdst1=pdst1
        psrlq           mm4, 32              // pframe2 = pframe3
        movd            esi, mm4
        add             esi, edx
        movd            mm5, esi             // pframe3 = pframe3 + iSrcPitch;
        psllq           mm5, 32
        por             mm5, mm4
        movq            i64pframe3_2, mm5
        
        movq            mm4, i64pframe2p_piFrameDiffs2p
        movq            mm5, mm4                // store
        psrlq           mm4, 32
        movd            edi, mm4                //pframe2p
                
		movq	        mm0, mm1				// bPlane0 = bPlane1;
		movq	        mm1, mm2				// bPlane1 = bPlane2;
		movq	        mm2, mm3				// bPlane2 = bPlane3;

		//if(y < src_height - 1)
		
		dec             cx
		dec             cx
		jg              metkaH
		movq	        mm3, mm1			//bPlane3=bPlane1
		jmp				end_metkaH
metkaH:
		 
		movq	        mm3, qword ptr[esi] //bPlane3 = *pframe3
		movq	        mm4, qword ptr[edi] //bY2p = *pframe2p
        movq            bY2p, mm4
end_metkaH:	
		//end if
		inc				cx          // recover ecx
		inc				cx
		add				edi, edx
        movd            mm4, edi    //pframe2p = pframe2p + iSrcPitch
        psllq           mm4, 32
        psllq           mm5, 32
        psrlq           mm5, 32
        por             mm5, mm4
        movq            i64pframe2p_piFrameDiffs2p, mm5 

  		movq	        mm4, i64FrameDiffValue1hih
		movq	        i64FrameDiffValue0hih, mm4
		movq	        mm4, i64FrameDiffValue1hil
		movq	        i64FrameDiffValue0hil, mm4
   		movq	        mm4, i64FrameDiffValue1loh
		movq	        i64FrameDiffValue0loh, mm4
   		movq	        mm4, i64FrameDiffValue1lol
		movq	        i64FrameDiffValue0lol, mm4
		movq	        mm4, i64InterlaceValue1hih
		movq	        i64InterlaceValue0hih, mm4
		movq	        mm4, i64InterlaceValue1hil
		movq	        i64InterlaceValue0hil, mm4
		movq	        mm4, i64InterlaceValue1loh
		movq	        i64InterlaceValue0loh, mm4
        movq	        mm4, i64InterlaceValue1lol
		movq	        i64InterlaceValue0lol, mm4
   		movq	        mm5, i64FrameDiffValue2hih
		movq	        i64FrameDiffValue1hih, mm5
		movq	        mm5, i64InterlaceValue2hih
		movq	        i64InterlaceValue1hih, mm5
		movq	        mm5, i64FrameDiffValue2hil
		movq	        i64FrameDiffValue1hil, mm5
		movq	        mm5, i64FrameDiffValue2loh
		movq	        i64FrameDiffValue1loh, mm5
		movq	        mm5, i64FrameDiffValue2lol
		movq	        i64FrameDiffValue1lol, mm5
		movq	        mm5, i64InterlaceValue2hil
		movq	        i64InterlaceValue1hil, mm5
		movq	        mm5, i64InterlaceValue2loh
		movq	        i64InterlaceValue1loh, mm5
		movq	        mm5, i64InterlaceValue2lol
		movq	        i64InterlaceValue1lol, mm5
		
		
             //if(y < src_height)
		pxor	        mm4, mm4			// mm4=0
		dec				cx
		jg				metkaH1
		
		movq	        i64FrameDiffValue2hih, mm4	
		movq	        i64InterlaceValue2hih, mm4
        movq	        i64InterlaceValue2hil, mm4
        movq	        i64InterlaceValue2loh, mm4
        movq	        i64InterlaceValue2lol, mm4	
        movq	        i64FrameDiffValue2hil, mm4
        movq	        i64FrameDiffValue2loh, mm4
        movq	        i64FrameDiffValue2lol, mm4	

		jmp				end_metkaH1

metkaH1:
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.

		
	    //Begin of iInterlaceValue2lo calculation	
		movq	     mm5, mm1
		punpcklbw    mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	     mm6, mm3
		punpcklbw    mm6, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 signed word (integer)
		psubsw	     mm5, mm6		//bPlane1 - bPlane3
		movq	     mm7, mm5		//copy bPlane1 - bPlane3

		PMULL_MMX(mm7, mm5, mm6)	//(bPlane1 - bPlane3)*(bPlane1 - bPlane3)

		movq	     mm5, mm7		//copy
		movq	     mm6, mm7
		movq	     mm7, i64EdgeDetect
		pmullw	     mm5, mm7		
        pmulhw	     mm6, mm7
        movq	     mm7, mm5		
		punpckhwd    mm5, mm6
        punpcklwd    mm7, mm6
		psrad	     mm5, 12			// >>12
		psrad	     mm7, 12			// >>12
		movq		 i64SavelumaLo, mm7 //save intermediate multiple result
		movq	     i64SavelumaHi, mm5 //save intermediate multiple result

		movq	      mm5, mm1
		punpcklbw     mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm6, mm2
		punpcklbw     mm6, mm4	// convert bPlane2 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm7, mm3
		punpcklbw     mm7, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 word (integer)
		psubsw	      mm5, mm6		//bPlane1 - bPlane2
		psubsw        mm7, mm6		//bPlane3 - bPlane2
		movq		  mm6, mm5 
		pmullw		  mm5, mm7 
		pmulhw		  mm6, mm7 
		movq		  mm7, mm5 
		punpckhwd	  mm5, mm6   //(bPlane1 - bPlane2)*(bPlane3 - bPlane2)
		punpcklwd	  mm7, mm6 
		psubd		  mm7, i64SavelumaLo
		psubd		  mm5, i64SavelumaHi

		movq	     mm6, mm5	    //store intermediate
		pslld	     mm5, 3			// multiple by
		pslld	     mm6, 1			// by
		paddd	     mm5, mm6		// 10

		movq	     mm6, mm7	    //store intermediate
		pslld	     mm7, 3			// multiple by
		pslld	     mm6, 1			// by
		paddd	     mm7, mm6		// 10
	
		movq	     i64InterlaceValue2loh, mm5
		movq	     i64InterlaceValue2lol, mm7
	//End of iInterlaceValue2lo calculation
		
	//Begin of iInterlaceValue2hi calculation		
		movq	     mm5, mm1
		punpckhbw    mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	     mm6, mm3
		punpckhbw    mm6, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 signed word (integer)
		psubsw	     mm5, mm6		//bPlane1 - bPlane3
		movq	     mm7, mm5		//copy bPlane1 - bPlane3

		PMULL_MMX(mm7, mm5, mm6)	//(bPlane1 - bPlane3)*(bPlane1 - bPlane3)

		movq	     mm5, mm7		//copy
		movq	     mm6, mm7
		movq	     mm7, i64EdgeDetect
		pmullw	     mm5, mm7		
        pmulhw	     mm6, mm7
        movq	     mm7, mm5		
		punpckhwd    mm5, mm6
        punpcklwd    mm7, mm6
		psrad	     mm5, 12			// >>12
		psrad	     mm7, 12			// >>12
		movq		 i64SavelumaLo, mm7 //save intermediate multiple result
		movq	     i64SavelumaHi, mm5 //save intermediate multiple result

		movq	      mm5, mm1
		punpckhbw     mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm6, mm2
		punpckhbw     mm6, mm4	// convert bPlane2 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm7, mm3
		punpckhbw     mm7, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 word (integer)
		psubsw	      mm5, mm6		//bPlane1 - bPlane2
		psubsw        mm7, mm6		//bPlane3 - bPlane2
		movq		  mm6, mm5 
		pmullw		  mm5, mm7 
		pmulhw		  mm6, mm7 
		movq		  mm7, mm5 
		punpckhwd	  mm5, mm6   //(bPlane1 - bPlane2)*(bPlane3 - bPlane2)
		punpcklwd	  mm7, mm6 
		psubd		  mm7, i64SavelumaLo
		psubd		  mm5, i64SavelumaHi

		movq	     mm6, mm5	    //store intermediate
		pslld	     mm5, 3			// multiple by
		pslld	     mm6, 1			// by
		paddd	     mm5, mm6		// 10

		movq	     mm6, mm7	    //store intermediate
		pslld	     mm7, 3			// multiple by
		pslld	     mm6, 1			// by
		paddd	     mm7, mm6		// 10
	
		movq	     i64InterlaceValue2hih, mm5
		movq	     i64InterlaceValue2hil, mm7

	//End of iInterlaceValue2hi calculation		


	//Begin of iFrameDiffValue2lo calculation
        movq          mm5, i64pframe2p_piFrameDiffs2p
        movq	      mm7, i64InterFrameAverage
        movd          esi, mm5                   // mov	 esi, piFrameDiffs2p    
		movq	      mm5, qword ptr[esi]	     //load *piFrameDiffs2p
        movq	      mm6, qword ptr[esi+8]
        movd          esi, mm7  
        movd          edi, mm5
        imul          edi, esi
        movd          mm4, edi                     // store low
        psrlq         mm5, 32
        movd          edi, mm5
        imul          edi, esi
        movd          mm5, edi
        psllq         mm5, 32
        por           mm5, mm4
        movq	      i64SavelumaLo, mm5        //store (iInterFrameAverage * *piFrameDiffs2p)
        movd          edi, mm6
        imul          edi, esi
        movd          mm4, edi                     // store low
        psrlq         mm6, 32
        movd          edi, mm6
        imul          edi, esi
        movd          mm6, edi
        psllq         mm6, 32
        por           mm6, mm4
		movq	      i64SavelumaHi, mm6        //store (iInterFrameAverage * *piFrameDiffs2p)
		
		dec           bl
		jz            lbln_0                  //if frame number n=0  go to lbln_0
		
        pxor	      mm4, mm4
		movq	      mm5, mm2
		punpcklbw     mm5, mm4        // convert bPlane2 (low order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		movq	      mm6, bY2p
		punpcklbw     mm6, mm4        //convert bY2p (low order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		
		psubsw	      mm5, mm6        // bPlane2-bY2p
		movq	      mm7, mm5	      // bPlane2-bY2p
		PMULL_MMX(mm7, mm5, mm6)      //result in mm7 = (bPlane2-bY2p)*(bPlane2-bY2p)
        movq	      mm5, i64InterFrameAverageRest
        movq          mm6, mm5        // copy
		pmullw	      mm5, mm7	      //iInterFrameAverageRest * (bPlane2-bY2p)*(bPlane2-bY2p)
        pmulhw	      mm6, mm7
        movq	      mm7, mm5
		punpckhwd     mm5, mm6
        punpcklwd     mm7, mm6
  		movq	      mm6, mm7	    //copy intermediate
		pslld	      mm7, 3	    // multiple
		pslld	      mm6, 1	    // by	
		paddd	      mm7, mm6	    // 10
		paddd         mm7, i64SavelumaLo
		movq	      mm6, mm5      //copy intermediate
		pslld	      mm5, 3	    // multiple
		pslld	      mm6, 1	    // by	
		paddd	      mm5, mm6	    // 10
		paddd         mm5, i64SavelumaHi     //iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest * 10*(bPlane2-bY2p)*(bPlane2-bY2p)
        jmp           End_lbln   	
lbln_0:				
        pxor          mm4, mm4
        movq          mm5, i64pframe2p_piFrameDiffs2p
        movq	      mm7, i64InterFrameAverageRest
        movd          esi, mm5                   // mov	 esi, piFrameDiffs2p    
		movq	      mm5, qword ptr[esi]	     //load *piFrameDiffs2p
        movq	      mm6, qword ptr[esi+8]
        punpcklwd     mm7, mm4
        movd          esi, mm7  
        movd          edi, mm5
        imul          edi, esi
        movd          mm4, edi                     // store low dword
        psrlq         mm5, 32
        movd          edi, mm5
        imul          edi, esi
        movd          mm7, edi
        psllq         mm7, 32
        por           mm7, mm4                      //store 2 low dwords(iInterFrameAverageRest * *piFrameDiffs2p)
        paddd         mm7, i64SavelumaLo
        movd          edi, mm6
        imul          edi, esi
        movd          mm4, edi                      // store low
        psrlq         mm6, 32
        movd          edi, mm6
        imul          edi, esi
        movd          mm5, edi
        psllq         mm5, 32
        por           mm5, mm4		                 //store 2 high dwords (iInterFrameAverageRest * *piFrameDiffs2p)
        paddd         mm5, i64SavelumaHi
End_lbln:
		inc			  bl
        movq          mm6, i64pframe2p_piFrameDiffs2p
        movd          esi, mm6                   // mov	 esi, piFrameDiffs2p
		psrad         mm5, 10       // >>10
		psrad         mm7, 10       // >>10
		movq	      qword ptr[esi], mm7
		movq	      qword ptr[esi+8], mm5
		movq	      i64FrameDiffValue2lol, mm7		//store	i64FrameDiffValue2lo
		movq	      i64FrameDiffValue2loh, mm5		//store	i64FrameDiffValue2lo
		

	//End of i64FrameDiffValue2lo calculation


	//Begin of i64FrameDiffValue2hi calculation
        movq	      mm5, qword ptr[esi+16]	     //load *piFrameDiffs2p
        movq	      mm6, qword ptr[esi+24]
        movq	      mm7, i64InterFrameAverage
        movd          esi, mm7  
        movd          edi, mm5
        imul          edi, esi
        movd          mm4, edi                     // store low
        psrlq         mm5, 32
        movd          edi, mm5
        imul          edi, esi
        movd          mm5, edi
        psllq         mm5, 32
        por           mm5, mm4
        movq	      i64SavelumaLo, mm5        //store (iInterFrameAverage * *piFrameDiffs2p)
        movd          edi, mm6
        imul          edi, esi
        movd          mm4, edi                     // store low
        psrlq         mm6, 32
        movd          edi, mm6
        imul          edi, esi
        movd          mm6, edi
        psllq         mm6, 32
        por           mm6, mm4
		movq	      i64SavelumaHi, mm6        //store (iInterFrameAverage * *piFrameDiffs2p)
	
		dec           bl
		jz            lbln_01                  //if n=0 go to lbln_0
        pxor	      mm4, mm4
		movq	      mm5, mm2
		punpckhbw     mm5, mm4        // convert bPlane2 (high order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		movq	      mm6, bY2p
		punpckhbw     mm6, mm4        //convert bY2p (high order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		
		psubsw	      mm5, mm6        // bPlane2-bY2p
		movq	      mm7, mm5	      // bPlane2-bY2p
		PMULL_MMX(mm7, mm5, mm6)      //result in mm7 = (bPlane2-bY2p)*(bPlane2-bY2p)
        movq	      mm5, i64InterFrameAverageRest
        movq          mm6, mm5        // copy
		pmullw	      mm5, mm7	      //iInterFrameAverageRest * (bPlane2-bY2p)*(bPlane2-bY2p)
        pmulhw	      mm6, mm7
        movq	      mm7, mm5
		punpckhwd     mm5, mm6
        punpcklwd     mm7, mm6
  		movq	      mm6, mm7	    //copy intermediate
		pslld	      mm7, 3	    // multiple
		pslld	      mm6, 1	    // by	
		paddd	      mm7, mm6	    // 10!
		paddd         mm7, i64SavelumaLo
		movq	      mm6, mm5      //copy intermediate
		pslld	      mm5, 3	    // multiple
		pslld	      mm6, 1	    // by	
		paddd	      mm5, mm6	    // 10!
		paddd         mm5, i64SavelumaHi     //iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest * 10*(bPlane2-bY2p)*(bPlane2-bY2p)
        jmp           End_lbln1   	
lbln_01:				
        pxor          mm4, mm4
        movq          mm5, i64pframe2p_piFrameDiffs2p
        movd          esi, mm5                   // mov	 esi, piFrameDiffs2p    
		movq	      mm5, qword ptr[esi+16]	     //load *piFrameDiffs2p
        movq	      mm6, qword ptr[esi+24]
        movq	      mm7, i64InterFrameAverageRest
        punpcklwd     mm7, mm4
        movd          esi, mm7  
        movd          edi, mm5
        imul          edi, esi
        movd          mm4, edi                     // store low dword
        psrlq         mm5, 32
        movd          edi, mm5
        imul          edi, esi
        movd          mm7, edi
        psllq         mm7, 32
        por           mm7, mm4                      //store 2 low dwords(iInterFrameAverageRest * *piFrameDiffs2p)
        paddd         mm7, i64SavelumaLo
        movd          edi, mm6
        imul          edi, esi
        movd          mm4, edi                      // store low
        psrlq         mm6, 32
        movd          edi, mm6
        imul          edi, esi
        movd          mm5, edi
        psllq         mm5, 32
        por           mm5, mm4		                 //store 2 high dwords (iInterFrameAverageRest * *piFrameDiffs2p)
        paddd         mm5, i64SavelumaHi
End_lbln1:
		inc			  bl
        movq          mm6, i64pframe2p_piFrameDiffs2p
        movd          esi, mm6                   // mov	 esi, piFrameDiffs2p
		psrad         mm5, 10       // >>10
		psrad         mm7, 10       // >>10
		movq	      qword ptr[esi+16], mm7
		movq	      qword ptr[esi+24], mm5

        movq	      i64FrameDiffValue2hil, mm7		//store	i64FrameDiffValue2hil
		movq	      i64FrameDiffValue2hih, mm5		//store	i64FrameDiffValue2hih
		add           esi, 32
        movd          mm5, esi
        psrlq         mm6, 32                       
        psllq         mm6, 32
        por           mm5, mm6
		movq		  i64pframe2p_piFrameDiffs2p, mm5		// piFrameDiffs2p++;
 
	//End of i64FrameDiffValue2hi calculation

         
end_metkaH1:
		inc             cx
        ror             ebx, 16
        test            bh, bh      
        jz              bh1                 //  if(y > 0)
        dec             bh
        ror             ebx, 16
		jmp     		nextH
bh1:        
        ror             ebx, 16
	//if (bBlend=true)		
		dec			bh		
		jz			goBlend

	// Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
        inc         bh
		mov			esi, ecx
        shl         esi, 16
		shr         esi, 16
		and			esi, 1
        mov         edi, ebx
        shr         edi, 16
		xor			esi, edi    //xor iOddEven
		
		jz			NeitherInterlaced //Odd lines ( )	

        //compute low order values of compare mask
		movq		mm6, i64InterlaceValue1loh
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0loh //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2loh //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold // > iThreshold
		movq		mm4, i64FrameDiffValue1loh
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0loh //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2loh //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm5, mm4	//store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm5, i64InterFrameLeaveThreshold
		por			mm4, mm5	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		mm6, mm4	// > iThreshold and >= iInterframeLeaveThresh
		movq		mm5, i64InterlaceValue1lol
		pslld		mm5, 1
		paddd		mm5, i64InterlaceValue0lol //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm5, i64InterlaceValue2lol //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm5, i64Threshold // > iThreshold
		movq		mm4, i64FrameDiffValue1lol
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0lol //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2lol //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	//store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm4, mm7	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		mm5, mm4	// > iThreshold and >= iInterframeLeaveThresh
		packssdw	mm5, mm6
		
		//compute high order values of compare mask
		
		movq		mm6, i64InterlaceValue1hih
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0hih	 // iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2hih	// iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold
		movq		mm4, i64FrameDiffValue1hih
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0hih	// iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2hih	// iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	// store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm7, mm4	// compare mask for >= iInterframeLeaveThresh
		pand		mm7, mm6	// > iThreshold && >= iInterframeLeaveThresh
		movq		i64SavelumaHi, mm7
		movq		mm6, i64InterlaceValue1hil
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0hil	 // iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2hil	// iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold
		movq		mm4, i64FrameDiffValue1hil
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0hil	// iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2hil	// iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	// store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm7, mm4	// compare mask for >= iInterframeLeaveThresh
		pand		mm7, mm6	// > iThreshold && >= iInterframeLeaveThresh
		packssdw	mm7, i64SavelumaHi
		packsswb	mm5, mm7

		movq		mm7, mm5 // store compare mask for 8 bytes

        movq		mm4, mm0
		pavgb		mm4, mm2	// (bPlane0 + bPlane2) / 2
		jmp			End_interlaced
				
goBlend:
		// Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
		inc			bh
        //compute low order values of compare mask
		movq		mm6, i64InterlaceValue1loh
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0loh //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2loh //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold // > iThreshold
		movq		mm4, i64FrameDiffValue1loh
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0loh //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2loh //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm5, mm4	//store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm5, i64InterFrameLeaveThreshold
		por			mm4, mm5	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		mm6, mm4	// > iThreshold and >= iInterframeLeaveThresh
		movq		mm5, i64InterlaceValue1lol
		pslld		mm5, 1
		paddd		mm5, i64InterlaceValue0lol //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm5, i64InterlaceValue2lol //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm5, i64Threshold // > iThreshold
		movq		mm4, i64FrameDiffValue1lol
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0lol //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2lol //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	//store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm4, mm7	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		mm5, mm4	// > iThreshold and >= iInterframeLeaveThresh
		packssdw	mm5, mm6
		
		//compute high order values of compare mask
		
		movq		mm6, i64InterlaceValue1hih
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0hih	 // iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2hih	// iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold
		movq		mm4, i64FrameDiffValue1hih
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0hih	// iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2hih	// iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	// store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm7, mm4	// compare mask for >= iInterframeLeaveThresh
		pand		mm7, mm6	// > iThreshold && >= iInterframeLeaveThresh
		movq		i64SavelumaHi, mm7
		movq		mm6, i64InterlaceValue1hil
		pslld		mm6, 1
		paddd		mm6, i64InterlaceValue0hil	 // iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		mm6, i64InterlaceValue2hil	// iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		mm6, i64Threshold
		movq		mm4, i64FrameDiffValue1hil
		pslld		mm4, 1
		paddd		mm4, i64FrameDiffValue0hil	// iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		mm4, i64FrameDiffValue2hil	// iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		mm7, mm4	// store
		pcmpeqd		mm4, i64InterFrameLeaveThreshold
		pcmpgtd		mm7, i64InterFrameLeaveThreshold
		por			mm7, mm4	// compare mask for >= iInterframeLeaveThresh
		pand		mm7, mm6	// > iThreshold && >= iInterframeLeaveThresh
		packssdw	mm7, i64SavelumaHi
		packsswb	mm5, mm7

		movq		mm7, mm5 // store compare mask for 8 bytes

		movq		mm4, mm0		// bPlane0
		pavgb		mm4, mm1		// bPlane1 
		movq		mm6, mm2		// bPlane2
		pavgb		mm6, mm1
		pavgb		mm4, mm6		// mm4=(bPlane0 + 2 * bPlane1 + bPlane2) / 4
		jmp			End_interlaced
NeitherInterlaced:
		pxor		mm7, mm7
End_interlaced:
		pand		mm4, mm7			// result of compare with "interlaced pixels" mask
        movq        mm5, i64pframe1_pdst1
        movq        mm6, mm5        // store
        movd		edi, mm5        // pdst1
		psrlq       mm5, 32
        movd        esi, mm5        // pframe1
        movq        mm5, i64widthYV12_dst_p
		pandn		mm7, qword ptr[esi]			// leave untouched noninterlaced pixels
		por			mm4, mm7             // combine compare results   
		movntq		qword ptr[edi], mm4 // *pdst1=*pframe1
        pxor        mm4, mm4
        movq        mm7, mm5            // store
        punpckldq   mm5, mm4            // load dst_p
        paddd       mm5, mm6            // pdst1 = pdst1 + iDstPitch
        movq        i64pframe1_pdst1, mm5
		//end	if

nextH:
		dec			cx
		jge			Loop_H
	//end loop for height
	add				eax, 8
    psrlq           mm7, 32             // load widthYV12
    movd            esi, mm7
	cmp				eax, esi
	jb				LoopW
// Compute remained 0-6 pixels of width////////////////////////////////////////////////////////////
    movq        mm6, i64iWidth_rest_src_w
    psrlq       mm6, 32
	movd		esi, mm6                // load iWidth_rest
	test 		esi, esi
	jz          End_frame
    //Last remained 2 or 4 or 6 bytes 
LoopW_rest:
        ror             ebx, 16
        mov             bh, 1                   // for check if current y = 0 
        ror             ebx, 16
        mov             esi, ecx
        shr             esi, 16
        xor             cx, cx
        add             ecx, esi                //restore in cx src_h
		movq		    mm4, i64pframe_pframeprev //pframe - in high, pframeprev - low order dwords
		movq		    mm5, mm4
		psrlq		    mm5, 32
		movd		    esi, mm5			//mov	esi, pframe  -	current row, current frame
		add			    esi, eax			// pframe3 = pframe + x;
        mov             edi, ecx            // store ecx
        xor             ecx, ecx    
        mov             cx, word ptr[esi]
        movd		    mm3, ecx            //in mm3 - 2 bytes from current line, current frame (bPlane3 = *pframe3)
		movd            mm5, esi            //load pframe3        
        add			    esi, edx			// 
        mov             cx, word ptr[esi]
        movd            mm2, ecx            // in mm2 - 2 bytes from next line, current frame (bPlane2 = *pframe2)
		movd            mm6, esi            // pframe2
        mov             ecx, edi            // restore ecx
        psllq           mm5, 32
        por             mm5, mm6
        movq            i64pframe3_2, mm5   // pframe3 - high, pframe2 - in low order dwords
        movd            mm5, eax            //load x
        psllq           mm5, 32
        movd            mm6, eax
        por             mm5, mm6
   		movq		    mm6, mm4            //load with pframeprev
        psllq           mm6, 32
        movq            mm4, i64pdst_piFrameDiffs
        movq            mm7, mm4
        psrlq           mm7, 32             //load with pdst
        por             mm6, mm7
        paddd           mm6, mm5            // pframe2p = pframeprev + x - in high order dword
                                            // pdst1 = pdst + x          - in low order dword
        movq            i64pframe1_pdst1, mm6  //save pdst1, "pframe1" could be ignored 
        pxor            mm5, mm5
        movq            mm7, i64pframe2p_piFrameDiffs2p
        punpckldq       mm7, mm5            //piFrameDiffs2p
        punpckhdq       mm5, mm6            //pframe2p
        por             mm7, mm5
        movq            i64pframe2p_piFrameDiffs2p, mm7
        
        movd            mm6, eax            // load x (eax)
        mov             esi, ecx
        shr             esi, 16             // load src_h
        movd            mm5, esi
        movq            mm7, mm5
	    pmullw          mm5, mm6
	    pmulhw          mm7, mm6            
	    punpcklwd       mm5, mm7            // x*src_h
        paddd           mm5, mm4            // piFrameDiffs2p=piFrameDiffs + x * src_h - in low order dword
        pxor            mm4, mm4
        punpckldq       mm5, mm4
        movq            mm6, i64pframe2p_piFrameDiffs2p
        punpckhdq       mm4, mm6             // load pframe2p
        por             mm5, mm4    
        movq		    i64pframe2p_piFrameDiffs2p, mm5 // piFrameDiffs2p = piFrameDiffs + x * src_h - in low order dword
		
		pxor		    mm4, mm4
		movq		    i64FrameDiffValue2lol, mm4
		movq		    i64InterlaceValue2lol, mm4
		movq		    i64FrameDiffValue1lol, mm4
		movq		    i64InterlaceValue1lol, mm4
      
      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
 	  /*for(int y = 0; y <= src_height; y++)
      
		*/
Loop_H_rest:			
        movq            mm4, i64pframe3_2
        movq            mm6, mm4
        psllq           mm6, 32              // pframe2 - in high dword                                                                
        movq            mm5, i64pframe1_pdst1
        psllq           mm5, 32
        psrlq           mm5, 32              //load pdst1 in low dword
        por             mm6, mm5             // pframe1 = pframe2
        movq            i64pframe1_pdst1, mm6 // store pframe1 = pframe2, pdst1=pdst1
        psrlq           mm4, 32              // pframe2 = pframe3
        movd            esi, mm4
        add             esi, edx
        movd            mm5, esi             // pframe3 = pframe3 + iSrcPitch;
        psllq           mm5, 32
        por             mm5, mm4
        movq            i64pframe3_2, mm5
                
		movq	        mm0, mm1				// bPlane0 = bPlane1;
		movq	        mm1, mm2				// bPlane1 = bPlane2;
		movq	        mm2, mm3				// bPlane2 = bPlane3;

        movq            mm4, i64pframe2p_piFrameDiffs2p
        movq            mm5, mm4                // store
        psrlq           mm4, 32                 // load pframe2p

		//if(y < src_height - 1)
				
		dec             cx
		dec             cx
		jg              metkaH_rest
		movq	        mm3, mm1			//bPlane3=bPlane1
		jmp				end_metkaH_rest
metkaH_rest:
        mov             edi, ecx            // store ecx
        xor             ecx, ecx
        mov             cx, word ptr[esi]		 
		movd	        mm3, ecx            //bPlane3 = *pframe3
        movd            esi, mm4            //pframe2p
        mov             cx, word ptr[esi]
		movd	        mm6, ecx            //bY2p = *pframe2p
        movq            bY2p, mm6
        mov             ecx, edi            // restore ecx


end_metkaH_rest:	
		//end if
		inc				cx          // recover ecx
		inc				cx
        movd            esi, mm4    // load pframe2p
		add				esi, edx
        movd            mm4, esi    //pframe2p = pframe2p + iSrcPitch
        psllq           mm4, 32
        psllq           mm5, 32
        psrlq           mm5, 32
        por             mm5, mm4
        movq            i64pframe2p_piFrameDiffs2p, mm5 

        movq	        mm4, i64InterlaceValue1lol
		movq	        i64InterlaceValue0lol, mm4
		movq	        mm4, i64InterlaceValue2lol
		movq	        i64InterlaceValue1lol, mm4
   		movq	        mm4, i64FrameDiffValue1lol
		movq	        i64FrameDiffValue0lol, mm4
		movq	        mm4, i64FrameDiffValue2lol
		movq	        i64FrameDiffValue1lol, mm4
		
             //if(y < src_height)
        pxor	        mm4, mm4
		dec				cx
		jg				metkaH1_rest

		movq	        i64InterlaceValue2lol, mm4	
        movq	        i64FrameDiffValue2lol, mm4	
		jmp				end_metkaH1_rest
metkaH1_rest:
      
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
        //Begin of iInterlaceValue2lo calculation
		movq	     mm5, mm1
		punpcklbw    mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	     mm6, mm3
		punpcklbw    mm6, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 signed word (integer)
		psubsw	     mm5, mm6		//bPlane1 - bPlane3
		movq	     mm7, mm5		//copy bPlane1 - bPlane3
		PMULL_MMX(mm7, mm5, mm6)	//(bPlane1 - bPlane3)*(bPlane1 - bPlane3)
		movq	     mm5, mm7		//copy
		movq	     mm6, mm7
		movq	     mm7, i64EdgeDetect
		pmullw	     mm5, mm7		
        pmulhw	     mm6, mm7
        punpcklwd    mm5, mm6
		psrad	     mm5, 12			// >>12
		movq	     i64SavelumaLo, mm5 //save intermediate multiple result
		movq	      mm5, mm1
		punpcklbw     mm5, mm4	// convert bPlane1 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm6, mm2
		punpcklbw     mm6, mm4	// convert bPlane2 (low order bytes) from 8 unsigned byte to 4 word (integer)
		movq	      mm7, mm3
		punpcklbw     mm7, mm4	// convert bPlane3 (low order bytes) from 8 unsigned byte to 4 word (integer)
		psubsw	      mm5, mm6		//bPlane1 - bPlane2
		psubsw        mm7, mm6		//bPlane3 - bPlane2
		movq		  mm6, mm5 
		pmullw		  mm5, mm7 
		pmulhw		  mm6, mm7 
		punpcklwd	  mm5, mm6  //(bPlane1 - bPlane2)*(bPlane3 - bPlane2)
		psubd		  mm5, i64SavelumaLo
		movq	     mm6, mm5	    //store intermediate
		pslld	     mm5, 3			// multiple by
		pslld	     mm6, 1			// by
		paddd	     mm5, mm6		// 10
		movq	     i64InterlaceValue2lol, mm5
       //End of iInterlaceValue2lo calculation
		
///////////////////////////////////////////////////////////////////////////////////////////////

	//Begin of iFrameDiffValue2lo calculation
        pxor          mm4, mm4
        movq          mm5, i64pframe2p_piFrameDiffs2p
        movq	      mm7, i64InterFrameAverage
        movd          esi, mm5                   // mov	 esi, piFrameDiffs2p    
		movq	      mm5, qword ptr[esi]	     //load *piFrameDiffs2p
        punpcklwd     mm7, mm4
        movd          esi, mm7  
        movd          edi, mm5
        imul          edi, esi
        movd          mm4, edi                     // store low
        psrlq         mm5, 32
        movd          edi, mm5
        imul          edi, esi
        movd          mm5, edi
        psllq         mm5, 32
        por           mm5, mm4                      // iInterFrameAverage * *piFrameDiffs2p
		
		dec           bl
		jz            lbln_0_rest                  //if frame number n=0  go to lbln_0
        pxor	      mm4, mm4
		movq	      mm7, mm2
		punpcklbw     mm7, mm4        // convert bPlane2 (low order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		movq	      mm6, bY2p
		punpcklbw     mm6, mm4        //convert bY2p (low order bytes) from 8 unsigned byte to 4 word (unsigned integer)
		psubsw	      mm7, mm6        // bPlane2-bY2p
		movq	      mm6, mm7	      // bPlane2-bY2p
		PMULL_MMX(mm7, mm6, mm4)      //result in mm7 = (bPlane2-bY2p)*(bPlane2-bY2p)
        movq	      mm4, i64InterFrameAverageRest
        movq          mm6, mm4        // copy
		pmullw	      mm4, mm7	      //iInterFrameAverageRest * (bPlane2-bY2p)*(bPlane2-bY2p)
        pmulhw	      mm6, mm7
		punpcklwd     mm4, mm6
  		movq	      mm6, mm4	    //copy intermediate
		pslld	      mm4, 3	    // multiple
		pslld	      mm6, 1	    // by	
		paddd	      mm4, mm6	    // 10
		paddd         mm5, mm4      //iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest * 10*(bPlane2-bY2p)*(bPlane2-bY2p)
        jmp           End_lbln_rest   	

lbln_0_rest:				
        pxor          mm4, mm4
        movq          mm6, i64pframe2p_piFrameDiffs2p
        movq	      mm7, i64InterFrameAverageRest
        movd          esi, mm6                   // mov	 esi, piFrameDiffs2p    
		movq	      mm6, qword ptr[esi]	     //load *piFrameDiffs2p
        punpcklwd     mm7, mm4
        movd          esi, mm7  
        movd          edi, mm6
        imul          edi, esi
        movd          mm4, edi                     // store low dword
        psrlq         mm6, 32
        movd          edi, mm6
        imul          edi, esi
        movd          mm6, edi
        psllq         mm6, 32
        por           mm6, mm4                      //store 2 low dwords(iInterFrameAverageRest * *piFrameDiffs2p)
        paddd         mm5, mm6                      // iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest * *piFrameDiffs2p
End_lbln_rest:
        inc           bl
        movq          mm6, i64pframe2p_piFrameDiffs2p
        movd          esi, mm6
		mov			  edi, esi
		psrad         mm5, 10       // >>10
		movq	      qword ptr[edi], mm5
		movq	      i64FrameDiffValue2lol, mm5		//store	i64FrameDiffValue2lo
		add           esi, 8
        movd          mm5, esi                          // piFrameDiffs2p
        psrlq         mm6, 32                       
        psllq         mm6, 32
        por           mm5, mm6
		movq		  i64pframe2p_piFrameDiffs2p, mm5		// piFrameDiffs2p++;
	//End of i64FrameDiffValue2lo calculation

end_metkaH1_rest:
		inc             cx
        ror             ebx, 16
        test            bh, bh      
        jz              bh1_r                 //  if(y > 0)
        dec             bh
        ror             ebx, 16
		jmp     		nextH_rest
bh1_r:        
        ror             ebx, 16
	//if (bBlend=true)		
		dec			    bh		
		jz			    goBlend_rest

	// Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
		mov			    esi, ecx
        shl             esi, 16
		shr             esi, 16
		and			    esi, 1
        mov             edi, ebx
        shr             edi, 16
		xor			    esi, edi    //xor iOddEven
		
		jz			    NeitherInterlaced_rest //Odd lines ( )	

		//compute low order values of compare mask
		movq		    mm6, i64InterlaceValue1lol
		pslld	    	mm6, 1
		paddd		    mm5, i64InterlaceValue0lol //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		    mm5, i64InterlaceValue2lol //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		    mm5, i64Threshold // > iThreshold
		movq		    mm6, i64FrameDiffValue1lol
		pslld		    mm6, 1
		paddd		    mm4, i64FrameDiffValue0lol //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		    mm4, i64FrameDiffValue2lol //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		    mm6, mm4	//store
		pcmpeqd		    mm4, i64InterFrameLeaveThreshold
		pcmpgtd		    mm6, i64InterFrameLeaveThreshold
		por			    mm4, mm6	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		    mm5, mm4	// > iThreshold and >= iInterframeLeaveThresh
        pxor            mm4, mm4
		packssdw	    mm5, mm4
		packsswb	    mm5, mm4
		movq		    mm7, mm5 // store compare mask for 2 bytes

        movq	    	mm4, mm0
		pavgb	    	mm4, mm2	// (bPlane0 + bPlane2) / 2
		jmp		    	End_interlaced_rest
				
goBlend_rest:
		// Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
		movq		    mm6, i64InterlaceValue1lol
		pslld	    	mm6, 1
		paddd		    mm5, i64InterlaceValue0lol //iInterlaceValue0 + 2 * iInterlaceValue1
		paddd		    mm5, i64InterlaceValue2lol //iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 
		pcmpgtd		    mm5, i64Threshold // > iThreshold
		movq		    mm6, i64FrameDiffValue1lol
		pslld		    mm6, 1
		paddd		    mm4, i64FrameDiffValue0lol //iFrameDiffValue0 + 2 * iFrameDiffValue1
		paddd		    mm4, i64FrameDiffValue2lol //iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 
		movq		    mm6, mm4	//store
		pcmpeqd		    mm4, i64InterFrameLeaveThreshold
		pcmpgtd		    mm6, i64InterFrameLeaveThreshold
		por			    mm4, mm6	//in mm4 - compare for >= iInterframeLeaveThresh
		pand		    mm5, mm4	// > iThreshold and >= iInterframeLeaveThresh
        pxor            mm4, mm4
		packssdw	    mm5, mm4
		packsswb	    mm5, mm4
		movq		    mm7, mm5 // store compare mask for 2 bytes

		movq		    mm4, mm0		// bPlane0
		pavgb		    mm4, mm1		// bPlane1 
		movq		    mm6, mm2		// bPlane2
		pavgb		    mm6, mm1
		pavgb		    mm4, mm6		// mm4=(bPlane0 + 2 * bPlane1 + bPlane2) / 4
		jmp			    End_interlaced_rest
NeitherInterlaced_rest:
		pxor		    mm7, mm7
End_interlaced_rest:
        inc             bh
		pand		    mm4, mm7			// result of compare with "interlaced pixels" mask
        movq            mm5, i64pframe1_pdst1
        movq            mm6, mm5        // store
		psrlq           mm5, 32
        movd            esi, mm5        // pframe1
        mov             edi, ecx        // store    ecx
        xor             ecx, ecx
		mov             cx,  word ptr[esi]
        movd            mm5, ecx
        pandn		    mm7, mm5			// leave untouched noninterlaced pixels
		por			    mm4, mm7             // combine compare results 
        movq            mm5, i64widthYV12_dst_p
        movd            ecx, mm4
        movd		    esi, mm6            // pdst1
		mov 		    word ptr[esi], cx  // *pdst1=*pframe1
        pxor            mm4, mm4
        movq            mm7, mm5            // store
        punpckldq       mm5, mm4            // load dst_p
        paddd           mm5, mm6            // pdst1 = pdst1 + iDstPitch
        movq            i64pframe1_pdst1, mm5
        mov             ecx, edi            // restore    ecx
		//end	if
nextH_rest:
		dec			cx
		jge			Loop_H_rest
	//end loop for src_height
   	add				eax, 2
    movq            mm6, i64iWidth_rest_src_w
	movd		    esi, mm6                // load src_w
	cmp 		    eax, esi
	jb				LoopW_rest
End_frame:
    nop

 	}
      return 0;
}


int __stdcall Filter::ProcessFrame_YV12(int n, byte* pframe, byte* pframeprev, byte* pdst, int* piFrameDiffs, const int src_w, int src_h, int  dst_p, int src_p)
{

    byte bPlane0, bPlane1 = 0, bPlane2, bPlane3;
    int iY2p;
    byte* pframe1;
    byte* pframe2;
    byte* pframe3;
    byte* pdst1;
    byte* pframe2p;
    int *piFrameDiffs2p;
    int iInterlaceValue0, iInterlaceValue1, iInterlaceValue2;
    int iFrameDiffValue0, iFrameDiffValue1, iFrameDiffValue2;

        //extern iThreshold, iEdgeDetect, iInterFrameAverage, iInterFrameAverageRest, iInterFrameLeaveThreshold, n, m_fDoubleRate, bBlend, bShowDeinterlacedAreaOnly;
    
    int iOddEven;
    if(Filter::m_fDoubleRate)
		iOddEven = (n & 1);
    else
      iOddEven = 1;
  // Do some area based deinterlacing on the Y plane.
    for(int x = 0; x < src_w; x++)
    {
      pframe3 = pframe + x;
      bPlane3 = *pframe3;
      pframe2 = pframe3 + src_p;
      bPlane2 = *pframe2;

      pframe2p = pframeprev + x;
      piFrameDiffs2p = piFrameDiffs + x * src_h;

      pdst1 = pdst + x;
      iInterlaceValue1 = iInterlaceValue2 = 0;
      iFrameDiffValue1 = iFrameDiffValue2 = 0;

      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
      for(int y = 0; y <= src_h; y++)
      {
        pframe1 = pframe2;
        pframe2 = pframe3;
        pframe3 = pframe3 + src_p;
        bPlane0 = bPlane1;
        bPlane1 = bPlane2;
        bPlane2 = bPlane3;
        if(y < src_h - 1)
        {
          bPlane3 = *pframe3;
          iY2p = *pframe2p;
        }
        else
        {
          bPlane3 = bPlane1;
        }
        pframe2p = pframe2p + src_p;

        iInterlaceValue0 = iInterlaceValue1;
        iInterlaceValue1 = iInterlaceValue2;
        iFrameDiffValue0 = iFrameDiffValue1;
        iFrameDiffValue1 = iFrameDiffValue2;

        if(y < src_h)
        {
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
          iInterlaceValue2 = 10 * ((bPlane1 - bPlane2) * (bPlane3 - bPlane2) - ((iEdgeDetect * (bPlane1 - bPlane3) * (bPlane1 - bPlane3)) >> 12));
          iFrameDiffValue2 = (iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest *
                              (n                       ?
                               10 * (bPlane2 - iY2p) * (bPlane2 - iY2p) :
                               *piFrameDiffs2p)) >> 10;
          *piFrameDiffs2p = iFrameDiffValue2;
          piFrameDiffs2p++;
        }
        else
        {
          iInterlaceValue2 = 0;
          iFrameDiffValue2 = 0;

        }

        if(y > 0)
        {
          // Get mean interlace value of 3 rows (i.e. 5 pixels are examined).
          // Middle row has twice the weight.
          if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold &&
             iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
          {
              // If interlaced
              if(bBlend)
              {
                // Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
                byte bPlaneDst;
                bPlaneDst = (bPlane0 + 2 * bPlane1 + bPlane2) / 4;
                *pdst1 = bPlaneDst;
              }
              else // Interpolate
              {
                // Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
                if((y & 1) ^ iOddEven)
                  *pdst1 = *pframe1;
                else
                {
                  byte bPlaneDst;
                  bPlaneDst = (bPlane0 + bPlane2) / 2;
                  *pdst1 = bPlaneDst;
                }
              }

              iDeinterlacedPixels++;
          }

          else
            // Neither is interlaced
              *pdst1 = *pframe1;

          if(bShowDeinterlacedAreaOnly)
	
          {
            int iY0 = (*pdst1)/4;

            if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold)
            {
              if(iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
               iY0 += 76;
              else
               iY0 += 103;
            }
            else
            {
              if(iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
               iY0 += 58;
              else iY0 += 47;
            }
           *pdst1 = (byte)iY0;
          }

          pdst1 = pdst1 + dst_p;
        }
      }
    }
    return 0;
}
    
  AVSValue __cdecl Create_Filter(AVSValue            args,
                               void*               user_data,
                               IScriptEnvironment* pEnv)
{

// The input could be separate fields or interlaced frames, and the
  // user could want to double the framerate or not.  To accomodate
  // all four of the possible combinations, we would need one of four
  // different types of input streams:  normal, Weave,
  // DoubleWeaveFields, and DoubleWeaveFrames.  [DoubleWeave()
  // handles differentiating between DoubleWeaveFields() and
  // DoubleFieldFrames(), so we only have to deal with a single
  // DoubleWeave() stream.]  The table below shows the correct
  // process for the two input formats and the two desired output
  // formats:
  //
  //           Display Rate (doublerate=)
  //        Frame (false)      Field (true)
  // Fields    Weave       DoubleWeave(Fields)
  // Frames     None       DoubleWeave(Frames)

  // Prepare all three types of input streams (only one will be used).

  PClip pChildNormal, pChildDoubleWeave;

    if(args[5].Defined())
    {
      if(args[5].AsBool() && !args[0].AsClip()->GetVideoInfo().IsTFF())
        pChildNormal = pEnv->Invoke("AssumeTFF", args[0].AsClip()).AsClip();
      else if (!args[5].AsBool() && !args[0].AsClip()->GetVideoInfo().IsBFF())
        pChildNormal = pEnv->Invoke("AssumeBFF", args[0].AsClip()).AsClip();
      else
        pChildNormal = args[0].AsClip();
    }
    else pChildNormal = args[0].AsClip();
    if (pChildNormal->GetVideoInfo().IsFieldBased())
      pChildNormal = pEnv->Invoke("Weave", pChildNormal).AsClip();
    pChildDoubleWeave = pEnv->Invoke("DoubleWeave", pChildNormal).AsClip();

  return new Filter(pChildNormal,
                    pChildDoubleWeave,
                    args[1].AsInt(24),
                    args[2].AsInt(20),
                    args[3].AsInt(35),
                    args[4].AsInt(80),
                    args[6].AsBool(false),
                    args[7].AsBool(false),
                    args[8].AsBool(false),
                    args[9].AsBool(false),
                    (float)args[10].AsFloat(0.0),
                    args[11].AsString("DeinterlaceSmooth.log"),
                    pEnv);
}

extern "C"
__declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* pEnv)
{
  pEnv->AddFunction("SmoothDeinterlace", "c[lacethresh]i[edgethresh]i[staticthresh]i[staticavg]i[tff]b[doublerate]b[blend]b[showlace]b[log]b[logpercent]f[logfile]s", Create_Filter, 0);
  return("`SmoothDeinterlace' plugin 1.5");
}
